// DiffractionImageRaxis.cpp
// Author: F. Remacle
// 25 January 2005
// 
// The bits of the DiffractionImage library which pertain to Raxis images
// images.

extern "C" {
#include "Bruker/FRM_LIB.H"
}

#include "DiffractionImage.h"

extern long* ReadFrameFile(char *frmfile, struct FRM_HEADER *frmhdr);

// if we are in development mode, then include things pertaining to
// assert.

#ifdef TEST_CODE
#define NDEBUG
#include <assert.h>
#endif

using namespace std;

namespace DI 
	{	
	void DiffractionImage::loadBruker()
		{
		long* largeImage;
		struct FRM_HEADER dummy;
		// read the header
		loadBrukerHeader();
		allocateImage();
#ifdef TEST_CODE
		assert(image != NULL);
#endif
		image=(unsigned int*) ReadFrameFile((char*)filename.c_str(), &dummy);
		}
    
	void DiffractionImage::loadBrukerHeader()
		{
		struct FRM_HEADER bruk_header;
		string mylines[96];
		char* detecttype;
		char* day;
		char* hour;
		char* dummy;
		float size;
		int input_file,version,lsblast,scanaxis;
		long bruk_format;
		long err;
		long header_bytes_length;
		
		detecttype=new char[15];
		// open the stream to the file
		input_file=frm_open_file((char*)filename.c_str(),0,&bruk_format,&err);
		if (err!=0 || (bruk_format!=86 && bruk_format!=100))
			{
			string msg="error opening Bruker file: ";
			msg+=filename;
			throw(DiffractionImageException(msg));
			}
		//loading header content from file
		err = frm_read_header(input_file,&bruk_header,&header_bytes_length);
    	err = frm_close_file(input_file);
		for(int i=0;i<96;i++)
			{
			dummy=new char[80];
			strncpy(dummy,bruk_header.line[i],80);
			dummy[79]='\0';
			mylines[i]=dummy;
			if(mylines[i].length() > 9)
				mylines[i]=mylines[i].substr(8,mylines[i].length()-1);
			else
				mylines[i]="empty field";
			}
		
		// loading the value into the corresponding fields
		//! raxis beam center coordinate is is pixels
		// therefore it needs to be converted in mm
		sscanf(mylines[1].c_str(),"%d",&version);
		sscanf(mylines[54].c_str(), "%f %f", &beamX, &beamY);
		sscanf(mylines[55].c_str(), "%f", &distance);
		distance *= 10;
		sscanf(mylines[63].c_str(), "%f", &wavelength);
		sscanf(mylines[40].c_str(), "%d", &width);
		sscanf(mylines[41].c_str(), "%d", &height);

		if(version>=8)
			{
			sscanf(mylines[78].c_str(),"%s %f",detecttype,&size);
			//what we actually get is the number of pixel per cm so need to
			//convert to pixel size in mm
			pixelX=10.0f/size;
			pixelY=10.0f/size;
			}
		else
			{
			if(version>6)
				{
				size=0.0f;
				sscanf(mylines[78].c_str(),"%s",detecttype);
				if(strcmp(detecttype,"MULTIWIRE")==0)
					size=10.f/47.5f;
				if(strcmp(detecttype,"CCD-PXL-2K")==0 ||
				   strcmp(detecttype,"CCD-PXL-L6000")==0)
					size=10.f/56.02f;
				if(strcmp(detecttype,"CCD-PXL-ARR")==0 || 
				   strcmp(detecttype,"CCD-PXL-L6500N")==0 ||
				   strcmp(detecttype,"CCD-PXL-l6500F")==0   )
					size=10.f/32.0f;
				if(strcmp(detecttype,"CCD-PXL-KAF1500")==0)
					size=10.f/51.2f;
				if(strcmp(detecttype,"CCD-PXL-KAF")==0 || 
				   strcmp(detecttype,"CCD-PXL-KAF2")==0 ||
				   strcmp(detecttype,"CCD-PXL-MSPD")==0 ||
				   strcmp(detecttype,"CCD-PXL")==0   )
					size=10.f/81.92f;
				if(strcmp(detecttype,"CCD-LDI")==0)
					size=10.f/83.333f;
				pixelX=size;
				pixelY=size;
				}
			}
		beamX*=pixelX;
		beamY*=pixelY;
		sscanf(mylines[33].c_str(), "%f", &oscStart);
		sscanf(mylines[34].c_str(), "%f", &size);
		oscEnd = oscStart+size;
		sscanf(mylines[65].c_str(), "%d",&scanaxis); 
		switch(scanaxis)
			{
			case 0:
				oscAxis="none";
				break;
			case 1:
				oscAxis="2 theta";
				break;
			case 2:
				oscAxis="omega";
				break;
			case 3:
				oscAxis="phi";
				break;
			case 4:
				oscAxis="chi";
				break;
			default:
				oscAxis="unknown";
				break;
			}
			
		sscanf(mylines[27].c_str(), "%f", &exposureTime);
		day=new char[20];			
		hour=new char[20];
		sscanf(mylines[26].c_str(), "%s %s",day,hour);
		date=day;
		date+=" ";
		date+=hour;
		sscanf(mylines[37].c_str(), "%f",&twoTheta);
			
		serialNo = "N/A";
		sscanf(mylines[42].c_str(), "%d",&lsblast);    
		if (lsblast==0)
			bigEndian = false;
		else
			bigEndian = true;
		
		//according to the documentation image is recorded from top to bottom
		//so when being read for viewing it is identical to the way it is mounted
		rapidAxis[0]=1;
		rapidAxis[1]=0;
		rapidAxis[2]=0;
		slowAxis[0]=0;
		slowAxis[1]=1;
		slowAxis[2]=0;
		
		format="BRUKER-FRAME";
		manufact="BRUKER";
		return;
		}

	};
