// DiffractionImageCalculations.cpp
// maintained by G.Winter
// 7th June 2004
// 
// The implementation of functionality for performing calculations with 
// the DiffractionImages. At the moment the following functions are 
// implemented:
// float[17] radial(void)
// int[17] histogram(void)
// $Id: DiffractionImageCalculations.cpp,v 1.32 2008/07/22 15:01:27 FR45 Exp $

#include "DiffractionImage.h"

namespace DI
	{
	/*int* DiffractionImage::profile(int fromX,int fromY,int toX, int toY)
		{
		}*/
		
	int* DiffractionImage::histogram(void)
		{
		int* results;
		results= new int[17];
		for (int i = 0; i < 17; i++)
			results[i]=0;
		int width = getWidth();
		int height = getHeight();
		unsigned short pixel;
		for (int i = 0; i < width * height; i++)
			{
			pixel = image[i];
			if (pixel & 0x8000)
				results[16] ++;
			else
				{
				if (pixel & 0x4000)
					results[15] ++;
				else 
					{
					if (pixel & 0x2000)
						results[14] ++;
					else 
						{
						if (pixel & 0x1000)
							results[13] ++;
						else 
							{
							if (pixel & 0x0800)
								results[12] ++;
							else 
								{
								if (pixel & 0x0400)
									results[11] ++;
								else
									{
									if (pixel & 0x0200)
										results[10] ++;
									else 
										{
										if (pixel & 0x0100)
											results[9] ++;
										else 
											{
											if (pixel & 0x0080)
												results[8] ++;
											else 
												{
												if (pixel & 0x0040)
													results[7] ++;
												else 
													{
													if (pixel & 0x0020)
														results[6] ++;
													else 
														{
														if (pixel & 0x0010)
															results[5] ++;
														else 
															{
															if (pixel & 0x0008)
																results[4] ++;
															else
																{
																if (pixel & 0x0004)
																	results[3] ++;
																else 
																	{
																	if (pixel & 0x0002)
																		results[2] ++;
																	else 
																		{
																		if (pixel & 0x0001)
																			results[1] ++;
																		else
																			results[0] ++;
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		return results;
		}

	int DiffractionImage::corner(void)
		{
		// does the image go to the corner (e.g. ADSC) or is it an 
		// inscribed circle (e.g. MAR 165; IP)
		int width = getWidth();
		int height = getHeight();
		int total = 0;
		int pix = 0;
		if (image == NULL)
			{
			std::cerr << "Image NULL" << std::endl;
			throw("Image is NULL");
			}
		// sample points in a circle which is centred on the detector centre
		// and has radius just larger than the detector size
		double rad = 0.25 * (0.5 * ((float) width + height) + sqrt(((float) width * width + height * height)));
		double pi = 4.0 * atan(1.0);
		// int rad = width / 2;
		for (double theta = 0.0; theta < (2 * pi); theta += (0.01 * pi))
			{
			int x = int((width / 2) + rad * cos(theta));
			int y = int((height / 2) + rad * sin(theta));
			// check point is inside the image
			if (x >= 0 && x < width && y >= 0 && y < width)
				{
				pix += 1;
				total += image[x + y * width];
				}
			}
		if (pix == 0)
			{
			std::cerr << "Error compting background" << std::endl;
			throw("Error in computing background!");
			}
		double average = ((float) total) / ((float) pix);
		if (average < 1.0)
			return 0;
		return 1;
		}

	float* DiffractionImage::radial(void)
		{
		return radial(16);
		}

	float* DiffractionImage::radial(int bins)
		{
		float* results;
		float* totals;
		int* counts;
		float dr, r;
		results=new float[bins];
		totals=new float[bins];
		counts=new int[bins];
		for (int i = 0; i < bins + 1; i++)
			{
			totals[i]=0;
			counts[i]=0;
			results[i]=0;
			}
		int width = getWidth();
		int height = getHeight();
		unsigned short pixel;
		int bin;
		dr = 0.5 * float(width) / float(bins);
		for (int i = 0; i < height; i++)
			{
			for (int j = 0; j < width; j++)
				{
				pixel = image[i * width + j];
				r = sqrt((i - 0.5 * height) * (i - 0.5 * height) + (j - 0.5 * width) * (j - 0.5 * width));
				bin = int(r / dr);
				if (bin < bins)
					{
					totals[bin] += pixel;
					counts[bin] ++;
					}
				else
					{
					totals[bins] += pixel;
					counts[bins] ++;
					}
				}
			}
		for (int i = 0; i < bins + 1; i++)
			{
			if (counts[i] != 0)
				results[i] = totals[i] / counts[i];
			else
				results[i] = 0;
			}
		return results;
		}

	// calculate how asymmetric the image is
	float* DiffractionImage::asymmetry(void)
		{
		float* results;
		double theta;
		double x, y;
		results=new float[18];
		int width = getWidth();
		int height = getHeight();
		for (int i = 0; i < 18; i++)
			results=0;
		for (int i = 0; i < 18; i++)
			{
			theta = ((double) i) * 4.0 * atan(1.0) / 18.0;
			x = cos(theta);
			y = sin(theta);
			double total_plus = 0;
			double total_minus = 0;
			for (int _i = 0; _i < height; _i ++)
				{
				int dy = _i - (height / 2);
				for (int _j = 0; _j < width; _j ++)
					{
					int dx = _j - (width / 2);
					double dot = dx * x + dy * y;
					if (dot > 0)
						total_plus += image[_i * width + _j];
					else
						total_minus += image[_i * width + _j];
					}
				}
			results[i] = (total_plus - total_minus) / (total_plus + total_minus);
			}
		return results;
		}
	
	//calculate an estimate of the gain by randoming taking 20x20 patch of the 
	//images and calculate the gain if the patch is "clear".
	double DiffractionImage::gain(void)
		{
		double bestgain;
		int nplane,npix,patchsize,ntries;
			nplane=0;
			npix=0;
		patchsize=30;
		bestgain=10000.0;
		ntries=1000;
			
		if(image==NULL)
			throw(DiffractionImageException("Image has not been loaded"));
		
		srand(time(NULL));
		
		for(int i=0;i<ntries;i++)
			{
			int xstart,ystart,det,sumxisq,sumyisq,sumxiyi,sumxizi,sumyizi;
			int sumxi,sumyi,sumzi;
			double a,b,c,gain,normabc,variance,mean;
			unsigned short zi;
			xstart= rand() % (width-patchsize);
			ystart= rand() % (height-patchsize);
			sumxi=0;
			sumyi=0;
			sumzi=0;
			det=0;
			sumxisq=0;
			sumyisq=0;
			sumxiyi=0;
			sumxizi=0;
			sumyizi=0;
			a=0.0;
			b=0.0;
			c=0.0;
			
			
			for(int xi=xstart;xi<xstart+patchsize;xi++)
				{
				for(int yi=ystart;yi<ystart+patchsize;yi++)
					{
					zi=image[(yi*width)+xi];
					//printf("     pixel %dx%d : %d \n",xi,yi,zi);
					//a few counts pixel means it is in a non acceptable zone to
					//estimate gain therefore we jump to the end of the loop to
					//force choosing another patch.
					if(zi<100)
						{
						xi=xstart+patchsize;
						yi=ystart+patchsize;
						}
					sumxisq=sumxisq+(xi*xi);
					sumyisq=sumyisq+(yi*yi);
					sumxiyi=sumxiyi+(xi*yi);
					sumxizi=sumxizi+(xi*zi);
					sumyizi=sumyizi+(yi*zi);
					sumxi=sumxi+xi;
					sumyi=sumyi+yi;
					sumzi=sumzi+zi;
					}
				}
				/*printf("     sum of all xi: %d \n",sumxi);
				printf("     sum of all yi: %d \n",sumyi);
				printf("     sum of all zi: %d \n",sumzi);
				printf("     sum of all xiyi: %d \n",sumxiyi);
				printf("     sum of all xizi: %d \n",sumxizi);
				printf("     sum of all yizi: %d \n",sumyizi);
				printf("     sum of all xi squared: %d \n",sumxisq);
				printf("     sum of all yi squared: %d \n",sumyisq);*/
			if(zi<100)
				det=0;
			else
				{
				det=(sumxisq*sumyisq)-(sumxiyi*sumxiyi);
				//printf("     determinant: %d \n",det);
				}
			if(det!=0)
				{
				a=(1.0/(1.0*det))*((sumyisq*sumxizi)-(sumxiyi*sumyizi));
				b=(1.0/(1.0*det))*((sumxisq*sumyizi)-(sumxiyi*sumxizi));
				c=((1.0*sumzi)/(patchsize*patchsize*1.0))-(a*((1.0*sumxi)/(patchsize*patchsize*1.0))+b*((1.0*sumyi)/(patchsize*patchsize*1.0)));
			
				//making (a,b,c) a unit length vector
				normabc=sqrt((a*a)+(b*b)+(c*c));
				c=c/normabc;
				//printf("     normal vector of fitted plane: %f %f %f \n",a/normabc,b/normabc,c);
				}
				
			//if the angle is not quite vertical or if the average intensity is 
			//clearly showing a spot, drop the patch, if the sum of intensity is
			//zero it means we probably are in the blank area of a scanned image
			//plate. If the patch contains any pixel which is white then drop
			//the patch.
			
			//For the information cos(1deg) ~ 0.9998477, so we allow the normal 
			//vector of the plane to be at most off the vertical by one degree.
			
			if(c<0.9998477 || zi<100) 
				{
				if(c<0.9998477 && !(zi<100))
					{
					nplane++;
					//printf("Plane fitting failed!\n");
					}
				else
					{
					npix++;
					//printf("Pixel count too low!\n");
					}
				if((nplane>4*ntries || npix > 4*ntries) && i==0)
					{
					printf("Plane fitting failed %d times\n",nplane);
					printf("Pixel count was too low %d times\n",npix);
					printf("impossible to find a proper area to estimate gain!\n");
					return -1;
					}
				else
					i--;
				}
			else
				{
				//printf("Succesfully fitted intensity plane!\n");
				mean=(1.0*sumzi)/(patchsize*patchsize*1.0);
				//calculating the variance.
				for(int xi=xstart;xi<xstart+patchsize;xi++)
					{
					for(int yi=ystart;yi<ystart+patchsize;yi++)
						{
						zi=image[(yi*width)+xi];
						variance=variance+((zi-mean)*(zi-mean));
						}
					}
					//printf("Patch %d \n",i);
				variance=variance/(patchsize*patchsize*1.0);
				//printf("     variance: %f\n",variance);
				//printf("     average pixel intensity: %f\n",mean);
				gain=variance/mean;
				//printf("gain for patch %d is %f\n",i,gain);
				if(gain<bestgain)
					{
					gaintopX=xstart;
					gaintopY=ystart;
					bestgain=gain;
					}
				}
			}
		//printf("bestgain found in %dx%d ->%dx%d\n",gaintopX,gaintopY,gaintopX+19,gaintopY+19);
		//printf("Plane fitting failed %d times\n",nplane);
		//printf("Pixel count was too low %d times\n",npix);
		return bestgain;
		}
	int DiffractionImage::gainX(void)
		{
		return gaintopX;
		}
		
	int DiffractionImage::gainY(void)
		{
		return gaintopY;
		}
		
	//This one is needed to be able to automagically generate a mask that covers
	// the back stop area and the arm of the backstop. It takes a beam center
	//coordinates as input if sensible, otherwise use header information
	//if header information is rubbish than we need to scan the image to try to
	//find a candidate area to start search from. It then returns the mask
	//parameters in xiaview format (-circlemask . . -polymask . . .)
	char* DiffractionImage::autoMask(int beamX, int beamY)
		{
		int startX,startY,nsteps,index;
		float average,variance,oldvariance,fitR,startR;
		double cosine,sine,mydist,magic,magicSimple;
		float* backstop; 
		char* returnval;
		char* circle;
		int* armfit;
		bool edge;
			
		returnval=new char[4000];
		circle=new char[50];
		if(isOriginalImg())
			{
			magic=0.333;
			magicSimple=0.8;
			}
		else
			{
			magic=0.555;
			magicSimple=1.3333;
			}
			
		//1 check beam centre coordinates provided
		if(beamX <=0 || beamY<=0)
			{	
			startX=(int) (getBeamX()/getPixelX());
			if(format.compare("SMV") == 0 && manufact.compare("ADSC") == 0)
				startY=getHeight() -((int) (getBeamY()/getPixelY()));
			else
				startY=(int) (getBeamY()/getPixelY());
			}
		else
			{
			startX=beamX;
			startY=beamY;
			}
		//2 if beam centre is still rubbish start from centre of image and search
		//  a blank zone.
		if(startX <=0 || startY<=0)
			{
			throw (DiffractionImageException("autoMask: no good beam centre coordinates"));
			//scan around to find a blank area, will do it later
				
			/*startX=getWidth()/2;
			startY=getHeight()/2;
			for(int i=startX-20; i<=startX+20;i++)
				{
				for(int j=startY-20; j<=startX+20;j++)
					{
					
					}
				}	
			*/
			}
		
		//check if the start pixel is blank, it is better if it is not
		while(image[startY*width+startX]==0)
			{
			startX+=5;
			startY+=5;
			}
			
		backstop=backstopDiscFit(startX,startY);
		if(backstop[0] == -1 && backstop[1] == -1 && backstop[2] == -1)
			{
			int* newS=correctCoord(startX,startY);
			backstop=backstopDiscFit(newS[0],newS[1]);	
			}
		
		if(fabs(backstop[0]-floor(backstop[0]))>0.5)
			startX= (int) ceil(backstop[0]);
		else
			startX= (int) floor(backstop[0]);
		if(fabs(backstop[1]-floor(backstop[1]))>0.5)
			startY= (int) ceil(backstop[1]);
		else
			startY= (int) floor(backstop[1]);
		startR=backstop[2];
		
		sprintf(circle,"-circlemask %d %d %f",startX,startY,startR);
		
		//we have generated the backstop area, now we go to look for the arm of
		//the backstop, keeping in mind that there might not be one!
		//we go round 360 degree from the circle centre and shoot rays each 2
		//degree.
		edge=false;
		float bestangle=-1;
		mydist=backstop[2];
		for(int i=0;i<360 && !edge;i++)
			{
			cosine=cos(M_PI*i/180.0);
			sine=sin(M_PI*i/180.0);
			if(!rayOnTileJoint(startX,startY,cosine,sine))
				{
				nsteps=ray(startX,startY,cosine,sine,magic);
				edge=(startX+nsteps*cosine>=width-1 || startY+nsteps*sine>=height-1) ||
						   (startX+nsteps*cosine<=0 || startY+nsteps*sine<=0);
				if(nsteps>mydist)
					{
					mydist=nsteps;
					bestangle=i;
					}
				}
			}
		//Firstly check if we did not find any arm!
		if(bestangle == -1 || fabs(mydist-startR)<5)
			return circle;
		else
			{
			//If by chance we already touched the edge, immediatly try a simple
			//fitting. If we are lucky this will finish here and will be quick.
			if(edge)
				{
				int* from = new int[2];
				int* to = new int[2];
				double test;
				from[0]=(int) floor(startX+(startR*0.99)*cos(M_PI*bestangle/180.0));
				from[1]=(int) floor(startY+(startR*0.99)*sin(M_PI*bestangle/180.0));
				to[0]=(int) ceil(startX+mydist*cos(M_PI*bestangle/180.0));
				to[1]=(int) ceil(startY+mydist*sin(M_PI*bestangle/180.0));
				armfit=simpleArmFit(from,to,bestangle,startR);
				test=sqrt(varianceQuad(armfit))/averageQuad(armfit);
				if(test < magicSimple)
					{
					sprintf(returnval,"%s -polymask %d %d %d %d %d %d %d %d",circle,
							armfit[0],armfit[1],armfit[2],armfit[3],
							armfit[4],armfit[5],armfit[6],armfit[7]);
					return returnval;
					}
				else
					{
					printf("-polymask %d %d %d %d %d %d %d %d\n",
						armfit[0],armfit[1],armfit[2],armfit[3],
						armfit[4],armfit[5],armfit[6],armfit[7]);
					//Simple fit got a quad that we are not satisfied with. Therefore,
					//we are going to look around slices to scan around the ray to find
					//the zone covering the arm of the backstop.
					armfit=complexArmFit(from,to,bestangle,startR);
					sprintf(returnval,"%s",circle);
					
					index=1;
					while(armfit[index*4]!=-1)
						{
						char* temp=new char[60];
						sprintf(temp," -polymask %d %d %d %d %d %d %d %d",
								armfit[4*(index-1)],armfit[4*(index-1)+1],
								armfit[4*(index-1)+2],armfit[4*(index-1)+3],
								armfit[4*index+2],armfit[4*index+3],
								armfit[4*index],armfit[4*index+1]);
						strcat(returnval,temp);
						index++;
						}
					return returnval;
					}
				}
			}
		
		mydist=startR;
		
		//This step is to attempt to detect a backstop arm that would not be
		//in line with the circle centre found earlier or to refine previous
		//solution.			
		//Second shoot a small range of rays to take the "whitest" ray possible
		//we hope that in this way our ray will coincide with the middle of 
		//the arm as best as possible.
		
		float scale=25;
		float aver=1000000;
		for(int i=0;i<4;i++)
			{
			float bestangle2=bestangle;
			int locstartX,locstartY;
			if(edge)
				{
				i=3;
				scale=0.02f;
				}
			else
				scale/=5.0f;
			printf("Automask: no straight arm found, refining angle: %f\n",bestangle2);
			locstartX= (int) floor(startX+startR*cos(M_PI*bestangle2/180.0));
			locstartY= (int) floor(startY+startR*sin(M_PI*bestangle2/180.0));
			for(float j=bestangle2-10*scale;j<bestangle2+10*scale;j+=scale)
				{
				int mysteps;
				bool myEdge=false;
				cosine=cos(M_PI*j/180.0);
				sine=sin(M_PI*j/180.0);
				if(!rayOnTileJoint(locstartX,locstartY,cosine,sine))
					{
					mysteps=ray(locstartX,locstartY,cosine,sine,magicSimple);
					myEdge=(locstartX+mysteps*cosine>=width-1 || locstartY+mysteps*sine>=height-1) ||
						   (locstartX+mysteps*cosine<=0 || locstartY+mysteps*sine<=0);
					if(myEdge)
						{
						average=rayAverage(locstartX,locstartY,cosine,sine,mysteps-1);
						if(average<aver)
							{
							bestangle2=j;
							aver=average;
							nsteps=mysteps;
							}
						if(!edge)
							{
							edge=true;
							bestangle=bestangle2;
							}
						}
					else
						{
						if(mysteps > nsteps)
							{
							nsteps=mysteps;
							bestangle2=j;
							}
						}
					}
				}			
			}
			
		//Our new start point is not the centre but a little inside the edge of 
		//the circle along the best angle so far.
		
		if(edge)
		//we touch the edge, now the hard part begins!, firstly find out which 
		//edge we reached.
			{
			printf("Automask: refinement succesfull with %f deg from %d %d\n",bestangle,startX,startY);
			//Ok now we now wich edge we have fallen onto, 
			//Firstly we are going to try a simple fit.
			int* from = new int[2];
			int* to = new int[2];
			double test;
			to[0]=(int) ceil(startX+(startR+nsteps)*cos(M_PI*bestangle/180.0));
			to[1]=(int) ceil(startY+(startR+nsteps)*sin(M_PI*bestangle/180.0));
			from[0]=(int) floor(startX+(startR*0.99)*cos(M_PI*bestangle/180.0));
			from[1]=(int) floor(startY+(startR*0.99)*sin(M_PI*bestangle/180.0));
			armfit=simpleArmFit(from,to,bestangle,startR);
			test=sqrt(varianceQuad(armfit))/averageQuad(armfit);
//			printf("test std-dev is %f times the average\n",test);
			if(test< magicSimple)
				{
				sprintf(returnval,"%s -polymask %d %d %d %d %d %d %d %d",circle,
						armfit[0],armfit[1],armfit[2],armfit[3],
						armfit[4],armfit[5],armfit[6],armfit[7]);
				return returnval;
				}
			else
				{
				printf("-polymask %d %d %d %d %d %d %d %d\n has been rejected (std-dev is too high)",
						armfit[0],armfit[1],armfit[2],armfit[3],
						armfit[4],armfit[5],armfit[6],armfit[7]);
				
				//Simple fit got a quad that we are not satisfied with. Therefore,
				//we are going to look around slices to scan around the ray to find
				//the zone covering the arm of the backstop.
				armfit=complexArmFit(from,to,bestangle,startR);
				sprintf(returnval,"%s",circle);
				
				index=1;
				while(armfit[index*4]!=-1 && index <41)
					{
					char* temp=new char[100];
					sprintf(temp," -polymask %d %d %d %d %d %d %d %d",
							armfit[4*(index-1)],armfit[4*(index-1)+1],
							armfit[4*(index-1)+2],armfit[4*(index-1)+3],
							armfit[(4*index)+2],armfit[(4*index)+3],
							armfit[4*index],armfit[(4*index)+1]);
					strcat(returnval,temp);
					index++;
					}
				return returnval;
				}
			}
		else
			return circle;
		}
	
	//Given a particular resolution, return the radius in mm of the 
	//corresponding resolution ring.
	float DiffractionImage::resolRingRadius(float resolution)
		{
		float theta;
		float R;
		theta=asin(wavelength/(2.0*resolution));
		R=distance*tan(2*theta);
		return R;
		}
		
	//Given a particular radius, return the resolution of the 
	//corresponding resolution ring.
	float DiffractionImage::resolOnRadius(float radius)
		{
		float theta;
		float res;
		theta=(atan(radius/distance))/2.0;
		res=wavelength/(2.0*sin(theta));
		return res;
		}
	};
