// DiffractionImageJpeg.cpp
// authors: G.Winter
// maintained by F.Remacle
// 17th December 2003
// 
// The Jpeg (old school) handling routines for the DiffractionImage
// module. In addition, all of the utility code for handling the image
// is contained herein.
// 
// 
// $Id: DiffractionImageJpeg.cpp,v 1.11 2008/05/07 11:00:51 fr45 Exp $

#include "DiffractionImage.h"

extern "C" 
	{
#include "jpeglib.h"
#include "jerror.h"
	}


namespace DI 
	{
    // I guess for this that I should recode all of the jpeg handling bits -
    // and this time make it possible to draw things onto the images too.
    // this could be fun. fortunately I have made the code above flexible
    // enough to allow this  
    // next some real meaty methods - involving JPEG images
    // the first will call the second, which will call the third
    // filename is the name of the output file

    void DiffractionImage::jpeg(string filename)
		{
		jpeg(filename, 85);
		}
   
    // make the image 1:1 with the original - i.e. zoom = 0 
    void DiffractionImage::jpeg(string filename,int quality)
		{
		jpeg(filename, quality, 0);
		}
    
    // zoom the image - if zoom > 0 enlarge, else shrink
    // -1 == 0 == 1 => 1:1
    // this will create a jpeg file named $filename, with the quality
    // and zoom configured manually. this will produce an 8 bit 
    // representation of the image, which will be calculated using
    // bit shifting as this is far cheaper than the alternative
    // this will call the FUNCTION writeJpegImage to do the "dirty 
    // work". 
    // n.b. the traditional way to display these images is as a negative - 
    // so areas of high intensity are darker.

    void DiffractionImage::jpeg(string filename,int quality,int zoom)
		{
		DiffractionImage::jpeg(filename, quality, zoom, true);
		}

    void DiffractionImage::jpeg(string filename,int quality,int zoom,bool optimise)
		{
		if ((zoom == 0) || (abs((float) zoom) == 1))
			writeJpegImage(image, width, height, quality,optimise, filename.c_str());
		else 
			{
			if (zoom > 1) // we wish to enbiggen the image
				{
				unsigned int * big;
				try { // allocate the image
					big = new unsigned int[width * height * zoom * zoom];
					} catch (bad_alloc & e) {throw (DiffractionImageException("not enough memory to allocate new image"));}
				if (big == NULL)
					throw (DiffractionImageException("not enough memory to allocate new image"));
				// this should be wrapped in a try, catch to enable the
				// allocated memory to be freed
				try {
					enlargeImage(image, width, height, zoom, big);
					writeJpegImage(big, width * zoom, height * zoom, quality,optimise, filename.c_str());
					} catch (DiffractionImageException e) 
							{
							// tidy up and rethrow the exception
							delete [] big;
							throw (e);
							}
				// tidy up anyway
				delete [] big;
				}
			else
				{
				// we wish to make the image nice and small
				unsigned int * small;
				int factor = -1 * zoom;
	    
				// note that there are two layers of error trapping here, 
				// since I am not sure exactly what will happen in C++ if an
				// allocation fails - better to be sure.
				try {
					small = new unsigned int[width * height / (factor * factor)];
					}catch (bad_alloc & e) {throw (DiffractionImageException("not enough memory to allocate new image"));}
				if (small == NULL)
					throw (DiffractionImageException("not enough memory to allocate new image"));
				// try and catch again
				try {
					reduceImage(image, width, height, factor, small);
					writeJpegImage(small, width / factor, height / factor, quality,optimise, filename.c_str());
					} catch (DiffractionImageException e)
							{
							delete [] small;
							throw (e);
							}
				delete [] small;
				}
			}
		}

	// this function is the one which performs most of the actual
	// spade work, and will be calling all of the above functions to
	// convert the unsigned short non-negative image into unsigned
	// char (byte) negative image.
	void writeJpegImage(unsigned int * image,int width,int height,int quality,bool optimise,const char * filename)
		{
		unsigned char * copy;
		// validate the input
		if (filename == NULL)
			throw (DiffractionImageException("NULL image filename"));
		if (image == NULL)
			throw (DiffractionImageException("NULL image"));
		if ((width <= 0) || (height <= 0))
			throw (DiffractionImageException("wrong size"));
		if ((quality < 5) || (quality > 100))
			throw (DiffractionImageException("wrong quality"));
	
		// if we haven't included JPEG support then abort now - but
		// raise an exception to make debugging more possible.
		//throw (DiffractionImageException("JPEG support not included define INCLUDE_JPEG"));

		// now we can get on with it!
		try {
			copy = new unsigned char[width * height];
			} catch (bad_alloc & e) {throw (DiffractionImageException("error allocating image copy"));}
		if (copy == NULL)
			throw (DiffractionImageException("error allocating image copy"));
		
		// I am not sure how these could fail, but wrap in a try/catch 
		// clause just to be sure.
		try	{
			downSample(image, width, height, optimise, copy);
			invertImage(copy, width, height);
			} catch (DiffractionImageException e)
					{
					delete [] copy;
					throw (e);
					}
		// now to create the actual jpeg image
		// most of this code is taken from the example in the 
		// jpeg-6b distribution.

		// note well - that this will write to a C file stream - so 
		// opening a channel to the standard output is not at this
		// moment an option - but could perhaps be so in the future.

		// perhaps writing the jpeg to memory will be useful too.
		struct jpeg_compress_struct cinfo;
		struct jpeg_error_mgr jerr;
		FILE * outfile;
		JSAMPROW row_pointer[1];
		int row_stride;
		outfile = fopen(filename, "wb");
		if (outfile == NULL)
			{
			delete [] copy;
			throw (DiffractionImageException("error opening output file"));
			}
	
		// this may need to be fixed, as the default behaviour on 
		// error is to exit()
		cinfo.err = jpeg_std_error(&jerr);
		jpeg_create_compress(&cinfo);
		jpeg_stdio_dest(&cinfo, outfile);
		cinfo.image_width = width;
		cinfo.image_height = height;
		cinfo.input_components = 1;
		cinfo.in_color_space = JCS_GRAYSCALE;
		jpeg_set_defaults(&cinfo);
		jpeg_set_quality(&cinfo, quality, TRUE);
		jpeg_start_compress(&cinfo, TRUE);
		row_stride = width;
	
		// iterate over the scanlines in the image and write the
		// output to the jpeg file.
		while (cinfo.next_scanline < cinfo.image_height)
			{
			row_pointer[0] = & (copy[cinfo.next_scanline * row_stride]);
			jpeg_write_scanlines(&cinfo, row_pointer, 1);
			}
		jpeg_finish_compress(&cinfo);
		fclose(outfile);
		jpeg_destroy_compress(&cinfo);
		delete [] copy;
		}

	};
