// DiffractionImageMAR.cpp
// maintained by G.Winter
// 10th June 2004
// 
// A method based on the marcvt application for reading all MAR research
// image formats, including image plates and CCD's.
// 
// This relies on marcvt in third_party. You have been warned!
// 
// 
// 
#include "DiffractionImage.h"

extern "C" 
	{
#define FULL_HEADER
#include "mar300_header.h"
#include "mar345_header.h"
#include "marccd_header.h"
#include "marcif_header.h"
  
extern void swaplong(void * data, int size);
extern int GetCIFHeader(char * program, char * filename, FILE * file);
extern MAR345_HEADER Getmar345Header(FILE * file);
extern MAR300_HEADER Getmar300Header(FILE * file);
extern MAR345_HEADER CIF2mar345Header(char * filename);
extern int Putmar345Header(int fd, MAR345_HEADER h);
extern int Putmar300Header(int fd, MAR300_HEADER h);
extern int get_pck(FILE * file, short int * data);
extern int get_pck32(FILE * file, long int * data);
extern int put_pck(unsigned short * data, int x, int y, int fd);

/* the following code is taken from the marcvt package
 * Copyright by:        Dr. Claudio Klein
 *                      X-ray Research GmbH, Hamburg*/
#define NONE 0
#define MAR345 1
#define PCK 2
#define CCD 9

int verbose=0;
int is_mar(char * f)
		{
		char mystr[256];
		int mybo;
		FILE* file = fopen(f, "rb");
		if (file == NULL)
			{
			fclose(file);
			return -1;
			}
		fseek( file, 1028, SEEK_SET );
		mystr[4] = '\0';
		mystr[6] = '\0';
		if ( fread( mystr, sizeof(char), 6, file) < 4 )
			{
			fclose(file);
			return NONE;
			}
		if ( strstr( mystr, "MMX" ) || strstr( mystr, "MARCCD" ) ) 
			{
			fclose(file);
			return CCD;
			}
		fseek( file, 0, SEEK_SET );
		if ( fread( &mybo, sizeof(int), 1, file) < 1) 
			{
			fclose(file);
			return NONE;
			}
		if ( mybo == 1234 ) 
			{
			fclose(file);
			return MAR345;
			}
		else 
			{
			if ( mybo >= 100 && mybo <= 5000 ) 
				{
				if ( strstr( f, ".pck") ) 
					{
					fclose(file);
					return PCK; 
					}
				}
			else
				{
				swaplong( &mybo, 1*sizeof(int) );
				if ( mybo == 1234 ) 
					{
					fclose(file);
					return MAR345;
					}
				else 
					{
					if ( mybo >= 100 && mybo <= 5000 )
						{
						if ( strstr( f, ".pck") ) 
							{
							fclose(file);
							return PCK; 
							}
						}
					}
				}
			}
		fclose(file);
		}
	}
//end of the code from marcvt, the following is DiffractionImage only
    

using namespace std;

namespace DI
	{

	void DiffractionImage::loadMAR()
		{
		short int* tmpimg;
		loadMARHeader();
		allocateImage();
		int type = is_mar((char *) filename.c_str());
		switch(type)
			{
			case CCD:
				loadRawImage(filename, 4096);
				break;
			default:
				FILE * file = fopen(filename.c_str(), "rb");
				if(width!=height)
					get_pck32(file,(long int*) image);
				else
					{
					tmpimg=new short int[width*height];
					get_pck(file, tmpimg);
					for(int i=0;i<width*height;i++)
						image[i]=tmpimg[i];
					free(tmpimg);
					}
				fclose(file);
			}
		}

	// this routine is based (very closely!) around 
	// mar_convert from the marcvt application
	//
	// http://www.marresearch.com/download.htm
	void DiffractionImage::loadMARHeader()
		{
		string msg="";
		MAR345_HEADER h345;
		MARCCD_HEADER hccd;
		int swap = 0;
		int type = is_mar((char *) filename.c_str());
		FILE * file;
		file = fopen(filename.c_str(), "rb");
		if (file == NULL)
			{
			msg="Error opening file :";
			msg+=filename;
			throw(DiffractionImageException(msg));
			}
		switch(type)
			{
			case NONE:
				msg=filename;
				msg+=" is not a proper MAR image";
				throw(DiffractionImageException(msg));
				break;
			case CCD:
				// read the CCD image header
				fseek(file, 1024, SEEK_SET);
				if (fread(&hccd, sizeof(MARCCD_HEADER), 1, file) < 1)
					{
					fclose(file);
					throw(DiffractionImageException("Error reading Mar CCD image header"));
					}
				if (hccd.header_byte_order < 1234 || hccd.header_byte_order > 4321)
					swaplong((void *) &hccd, sizeof(MARCCD_HEADER));
				if (hccd.header_byte_order < 1234 || hccd.header_byte_order > 4321)
					{
					fclose(file);
					throw(DiffractionImageException("Error byte swapping CCD header"));
					}
				// determine if the image need be byte swapped
#if (__linux || VMS || __ALPHA || _osf__)
				if (hccd.data_byte_order == 1234)
					swap = 1;
#elif (__sgi || SUN || __hpux)
				if (hccd.data_byte_order == 4321)
					swap = 1;
#endif
				// now get the information out of the header
				width = hccd.nfast;
				height = hccd.nslow;
				beamX = 0.001 * hccd.beam_x;
				beamY = 0.001 * hccd.beam_y;
				distance = 0.001 * hccd.xtal_to_detector;
				if(distance==0)
					{
					distance= 0.001*hccd.start_xtal_to_detector;
					if(distance==0)
						distance= 0.001*hccd.end_xtal_to_detector;
					}
				exposureTime = 0.001 * hccd.exposure_time;
					
				switch(hccd.rotation_axis)
					{
					case 0:
						oscStart = 0.001 * hccd.start_twotheta;
						oscEnd = 0.001 * hccd.end_twotheta;
						oscAxis="2 theta";
						break;
					case 1:
						oscStart = 0.001 * hccd.start_omega;
						oscEnd = 0.001 * hccd.end_omega;
						oscAxis="omega";
						break;
					case 2:
						oscStart = 0.001 * hccd.start_chi;
						oscEnd = 0.001 * hccd.end_chi;
						oscAxis="chi";
						break;
					case 3:
						oscStart = 0.001 * hccd.start_kappa;
						oscEnd = 0.001 * hccd.end_kappa;
						oscAxis="kappa";
						break;
					case 4:
						oscStart = 0.001 * hccd.start_phi;
						oscEnd = 0.001 * hccd.end_phi;
						oscAxis="phi";
						break;
					case 5:
						oscStart = 0.001 * hccd.start_delta;
						oscEnd = 0.001 * hccd.end_delta;
						oscAxis="delta";
						break;
					case 6:
						oscStart = 0.001 * hccd.start_gamma;
						oscEnd = 0.001 * hccd.end_gamma;
						oscAxis="gamma";
						break;
					default:
						throw(DiffractionImageException("MARCCD rotation axis out of range"));
						break;
					}
				if(oscStart==oscEnd || (oscStart!=oscEnd && oscEnd==0))
					oscEnd=oscStart+0.001*hccd.rotation_range;
					
				twoTheta=0.001 * hccd.start_twotheta;
				pixelX = 0.000001 * hccd.pixelsize_x;
				pixelY = 0.000001 * hccd.pixelsize_y;
				wavelength = 0.00001 * hccd.source_wavelength;
				// copy out the timestamp information
				char timestamp[32];
				memset(timestamp, 0, 32);	
				// look for a meaningful date...
				strncpy(timestamp, hccd.acquire_timestamp, 32);
				if (strlen(timestamp) == 0)
					strncpy(timestamp, hccd.header_timestamp, 32);
				if (strlen(timestamp) == 0)
					strncpy(timestamp, hccd.save_timestamp, 32);
				date = timestamp;
				serialNo = "N/A";
				format="MARCCD";
				fclose(file);
				break;
			case MAR345:
				// read a MAR 345 image header
				h345 = Getmar345Header(file);
				// next get the bits out of the MAR 345 header
				// this could be a MAR555 format! so we need to check the 
				// number of pixels!
				width = h345.size;
				height = h345.pixels/width;
				distance = h345.dist;
				wavelength = h345.wave;
				exposureTime = h345.time;
				if(h345.phiosc)
					{
					oscStart = h345.phibeg;
					oscEnd = h345.phiend;
					oscAxis="phi";
					}
				else
					{
					oscStart = h345.omebeg;
					oscEnd = h345.omeend;
					oscAxis="omega";
					}
				pixelX = h345.pixel_length*0.001;
				pixelY = h345.pixel_height*0.001;
				// these are stored in pixels so...
				beamX = h345.xcen * pixelX;
				beamY = h345.ycen * pixelY;
				// get the detector serial number
				char detector_sn[40];
				memset(detector_sn, 0, 40);
				sprintf(detector_sn, "%d", h345.scanner);
				date= h345.date;
				serialNo = detector_sn;
 				twoTheta=h345.theta;
				fclose(file);
				//The mar555 detector is not square but mar 345 image plates will
				//always be, therefore is the width and height are different
				//we know it is MAR image.
				format="MAR345";
				break;
			case PCK:
				// read a MAR 300 image header
				MAR300_HEADER h300;
				h300 = Getmar300Header(file);
				// next get the bits out of the MAR 300 header
				width = h300.pixels_x;
				height = h300.pixels_y;
				beamX = h300.centre_x*0.001;
				beamY = h300.centre_y*0.001;
				distance = h300.distance*0.001;
				wavelength = h300.lambda;
				exposureTime = h300.exptime_sec;
				oscStart = h300.phi_start;
				oscEnd = h300.phi_end;
				oscAxis = "phi";
				pixelX = h300.p_x*0.001;
				pixelY = h300.p_y*0.001;
				date = "N/A";
				serialNo = "N/A";
				fclose(file);
				format="MAR300";
			    twoTheta=0;
				break;
		// could add support for others MAR format image plates here too if any (CIF?)
		default:
			fclose(file);
			throw(DiffractionImageException("Unrecognised image type"));
      }
	  manufact="MAR";

  }
  
/*	void DiffractionImage::saveMAR()
		{
		FILE* file;
		int file2;
		int type = is_mar((char *) filename.c_str());
		switch(type)
			{
			case MAR345:
				MAR345_HEADER h345;
				file = fopen(filename.c_str(), "rb");
				h345 = Getmar345Header(file);
				fclose(file);
				// next set the bits into the MAR 345 header
				if (fabs(h345.xcen-beamX*1000) > 0.0001) 
					h345.xcen = beamX*1000;
				if (fabs(h345.ycen-beamY*1000) > 0.0001) 
					h345.ycen = beamY*1000;
				if (fabs(h345.dist-distance*1000) > 0.0001) 
					h345.dist = distance*1000;
				if (fabs(h345.phibeg-phiStart*1000) > 0.0001) 
					h345.phibeg = phiStart;
				if (fabs(h345.phiend-phiEnd*1000) > 0.0001)
					h345.phiend = phiEnd;
				file2 = open(filename.c_str(), MYWRITE);
				Putmar345Header(file2, h345);
				put_pck(image, width, height, file2);
				close(file2);
				break;
			case PCK:
				MAR300_HEADER h300;
				file = fopen(filename.c_str(), "rb");
				h300 = Getmar300Header(file);
				fclose(file);	
				// next set the bits into the MAR 300 header
				if (fabs(h300.centre_x-beamX*1000) > 0.0001) 
					h300.centre_x = beamX*1000;
				if (fabs(h300.centre_y-beamY*1000) > 0.0001) 
					h300.centre_y = beamY*1000;
				if (fabs(h300.distance-distance*1000) > 0.0001) 
					h300.distance = distance*1000;
				if (fabs(h300.lambda-wavelength*1000) > 0.0001) 
					h300.lambda = wavelength;
				if (fabs(h300.phi_start-phiStart*1000) > 0.0001) 
					h300.phi_start = phiStart;
				if (fabs(h300.phi_end-phiEnd*1000) > 0.0001) 
					h300.phi_end = phiEnd; 
				file2 = open(filename.c_str(), MYWRITE);
				Putmar300Header(file2, h300);
				put_pck(image, width, height, file2);
				close(file2);
				break;
			case CCD:
				MARCCD_HEADER hccd;
				file = fopen(filename.c_str(), "rb");
				if (file == NULL)
					throw(DiffractionImageException("Error opening file " + filename));
				// read the CCD image header
				fseek(file, 1024, SEEK_SET);
				if (fread(&hccd, sizeof(MARCCD_HEADER), 1, file) < 1)
					{
					fclose(file);
					throw(DiffractionImageException("Error reading Mar CCD image header"));
					}
				fclose(file);
				// now set the information back into the header structure
				if (fabs(hccd.beam_x-beamX*1000) > 0.0001) 
					hccd.beam_x = int(1000 * beamX);
				if (fabs(hccd.beam_y-beamY*1000) > 0.0001)
					hccd.beam_y = int(1000 * beamY);
				if (fabs(hccd.xtal_to_detector-distance*1000) > 0.0001) 
					hccd.xtal_to_detector = int(1000 * distance);
				if (fabs(hccd.source_wavelength-wavelength*1000) > 0.0001)
					hccd.source_wavelength = int(1000000 * wavelength);
				if (hccd.header_byte_order < 1234 || hccd.header_byte_order > 4321)
					swaplong((void *) &hccd, sizeof(MARCCD_HEADER));
				if (hccd.header_byte_order < 1234 || hccd.header_byte_order > 4321)
					{
					fclose(file);
					throw(DiffractionImageException("Error byte swapping CCD header"));
					}

				//write back with changes applied
				file = fopen(filename.c_str(), "wb");
				if (file == NULL)
					throw(DiffractionImageException("Error opening file " + filename));

				// write the CCD image header
				fseek(file, 1024, SEEK_SET);
				if (fwrite(&hccd, sizeof(MARCCD_HEADER), 1, file) < 1)
					{
					fclose(file);
					throw(DiffractionImageException("Error reading Mar CCD image header"));
					}
				fclose(file);
				// and copy back the image content into the file.
				file2 = open(filename.c_str(), MYWRITE);
				put_pck(image, width, height, file2);
				close(file2);
				break;
			default:
				throw(DiffractionImageException("Unrecognised image type"));
			}
		}*/

	};
