// DiffractionImageADSC.cpp
// maintained by G.Winter
// 17th December 2003
// 
// The bits of the DiffractionImage library which pertain to ADSC
// images.
// 
// 
// 

// C++ things
#include <iostream>
#include <fstream>
#include <algorithm>
#include <map>
#include <string>

// core C things - required for using open() and read() below
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#if defined _MSC_VER
#include <io.h>
#endif

#include "DiffractionImage.h"
// if we are in development mode, then include things pertaining to assert.

using namespace std;

namespace DI 
	{

	// read in an ADSC image - this will read the header first, then 
	// reopen the file to read the unsigned shorts out - byte 
	// swapping if needed. NOTE WELL - this is hacky as there is 
	// an assumption here that the images will be represented as
	// unsigned shorts, which could well be wrong!
	void DiffractionImage::loadSaturn()
		{
		loadSaturnHeader();  // read the header
		allocateImage(); // allocate the image structure
#ifdef TEST_CODE
		assert(image != NULL);
#endif
		// after this stage, the rest of the image is just a big pile
		// of unsigned shorts - another chance to share code! :o)
		// this routine will be defined in DiffractionImageRaw.cpp
		loadRawImage(filename, originalHeaderLength);
		}
  
	void DiffractionImage::loadSaturnHeader()
		{
		float tempx, tempy;
		string temp,token;
		char* temp1;
		char* temp2;
		char* temp3;
		char* tempvalues;
		int length, posDist, pos2T, count;
		//We only can be here if the loadADSC noticed it is a saturn image
		//Therefore we know that all keys/values have already been read in
		//we just need to populate the diffraction image values with the correct
		//keys.
		if (headerMap.find("header_bytes") == headerMap.end())
			throw(DiffractionImageException("header_bytes not found in header"));

		ASSERT_IN_HEADER("size1");
		ASSERT_IN_HEADER("size2");
		ASSERT_IN_HEADER("byte_order");
		ASSERT_IN_HEADER("scan_wavelength");
		ASSERT_IN_HEADER("ccd_spatial_beam_position");
		ASSERT_IN_HEADER("ccd_detector_size");
		ASSERT_IN_HEADER("ccd_gonio_values");
		ASSERT_IN_HEADER("ccd_gonio_names");
		ASSERT_IN_HEADER("rotation");

   		if (headerMap["byte_order"].compare("big_endian")==0)
			bigEndian = true;
		else
			bigEndian = false;
		// if we get here then everything we want is in the header
		// now fetch it all out into the header data structure
		// this will also be done via macro, with the following prototype
		GET_FROM_HEADER("size1", "%d", width);
		GET_FROM_HEADER("size2", "%d", height);
		GET_FROM_HEADER("scan_wavelength", "%f", wavelength);
		
		sscanf(headerMap["header_bytes"].c_str(), "%d",&originalHeaderLength);
		sscanf(headerMap["ccd_detector_size"].c_str(), "%f %f", &tempx,&tempy);
		//NB, that is the beam position in pixels! not in mm
		sscanf(headerMap["ccd_spatial_beam_position"].c_str(), "%f %f", &beamX, &beamY);
		//Therefore we need to do some calculations:
		pixelX=tempx/(width*1.0);
		pixelY=tempy/(height*1.0);
		beamX=beamX*pixelX;
		beamY=beamY*pixelY;
		exposureTime=0;
		sscanf(headerMap["rotation"].c_str(), "%f %f", &oscStart,&oscEnd);
		oscAxis=headerMap["rotation_axis_name"];
		length = headerMap["ccd_gonio_names"].length();	
		temp = headerMap["ccd_gonio_names"];
		transform(temp.begin(), temp.end(), temp.begin(), 
			  (int(*)(int))tolower);

		temp1=new char[length + 1];
		memset(temp1, 0, length + 1);
		strncpy(temp1, temp.c_str(), length);
		
  		temp2 = strtok(temp1," ");
		count=0;
  		while (temp2 != NULL)
		  {
		    token=temp2;
		    if(token.compare("distance") == 0 || token.compare("dist") == 0)
		      posDist = count;
		    if(token.compare("2theta") == 0 || token.compare("twotheta") == 0)
		      pos2T = count;
		    temp2 = strtok(NULL, " ");
		    count++;
		  }	

		delete [] temp1;

		length=headerMap["ccd_gonio_values"].length();
		tempvalues=new char[length + 1];
		memset(tempvalues, 0, length + 1);
		strncpy(tempvalues, headerMap["ccd_gonio_values"].c_str(), length);
		temp3 = strtok(tempvalues," ");
		count=0;
  		while (temp3 != NULL)
		  {
		    if(count==posDist)
		      sscanf(temp3,"%f",&distance);
		    if(count==pos2T)
		      sscanf(temp3,"%f",&twoTheta);
		    temp3 = strtok(NULL, " ");
		    count++;
		  }

		delete [] tempvalues;


		// get the detector serial number out
		if (headerMap.find("ccd_serial_number") != headerMap.end())
		  serialNo = headerMap["ccd_serial_number"];
		else
		  serialNo="N/A";
		if (headerMap.find("dtrek_date_time") != headerMap.end())
		  date=headerMap["dtrek_date_time"];
		else
		  date="N/A";
		sscanf(headerMap["ccd_detector_vectors"].c_str(), "%f %f %f %f %f %f",
		       &rapidAxis[0],&rapidAxis[1],&rapidAxis[2],
		       &slowAxis[0],&slowAxis[1],&slowAxis[2]);
		format="SMV";
		manufact="RIGAKU";
		}

	};
