#include <stdlib.h>
#include <stdio.h>
#include "DiffractionImage.h"
#include "Peak.h"
using namespace DI;

void printPeaks(char* filename, float threshold=-1)
	{
	PeakList* mypeaks;
	string Myfilename(filename);
	DiffractionImage* diff=new DiffractionImage(filename);
	printf("Searching for peaks...");
	if(threshold!=-1)
		{
		mypeaks=new PeakList();
		mypeaks->find(diff,100000,threshold);
		}
	else
		mypeaks=new PeakList(diff);
	printf("\nDone!\n %d peaks found\n\n",mypeaks->length());
	for(int i=0;i<mypeaks->length();i++)
		{
		Peak onepeak=mypeaks->get(i);
		printf("Peak %d = x:%f  y:%f  phi:%f  intensity:%f\n",i+1,onepeak.x,onepeak.y,onepeak.phi,onepeak.intensity);
		}
	}
	
void printHist (char* filename, int min, int max, int width)
	{
	PeakList* mypeaks;
	int* histogram;
	Peak onepeak;
	int nhist;
	int currenthist;
	float maxintensity;
	string Myfilename(filename);
	DiffractionImage* diff=new DiffractionImage(filename);
	printf("Searching for peaks...");
	mypeaks=new PeakList();
	mypeaks->find(diff,100000,1.0f*min);
	printf("\nDone!\n",mypeaks->length());
	onepeak=mypeaks->get(0);
	nhist=0;
	while(max-(nhist*width) > 0)
		nhist++;
	histogram=new int[nhist];
	currenthist=nhist-1;
	for(int i=0;i<mypeaks->length();i++)
		{
		onepeak=mypeaks->get(i);
		if(currenthist*width > onepeak.intensity)
			currenthist--;
		if(onepeak.intensity <= max)
			histogram[currenthist]+=1;
		}
	for(int i=0;i<nhist; i++)
		printf("%d -> %d : %d\n",i*width,(i+1)*width,histogram[i]);
	}

int main (int argc, char *argv[])
	{
	if(argc!=2 && argc!=4 && argc!=6)
		{
		printf("Wrong Number of arguments!\n\n");
		printf("Usage is  'printpeaks [-th <intensity_threshold>] <filename>' \n");
		printf("   or     'printpeaks [-hist <min> <max> <width>] <filename>' \n");
		return -1;
		}
	else
		{
		if(argc==2)
			printPeaks(argv[1]);
		else
			{
			if(strcmp(argv[1],"-th")==0)
				printPeaks(argv[3],atof(argv[2]));
			else
				{
				if(strcmp(argv[1],"-hist")==0)
					printHist(argv[5],atoi(argv[2]),atoi(argv[3]),atoi(argv[4]));
				else
					{
					printf("Unknown keyword!\n\n");
					printf("Usage is  'printpeaks [-th <intensity_threshold>] <filename>' \n");
					return -1;
					}
				}
			}
		return 0;
		}
	}
